/**
 * 
 */
package org.processmining.plugins.petrinet.analysis;

import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import javax.swing.JOptionPane;
import javax.swing.JPanel;

import org.deckfour.uitopia.api.event.TaskListener.InteractionResult;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.connections.ConnectionCannotBeObtained;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.plugin.annotations.PluginLevel;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.models.connections.petrinets.behavioral.InitialMarkingConnection;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.models.semantics.petrinet.PetrinetSemantics;
import org.processmining.models.semantics.petrinet.impl.PetrinetSemanticsFactory;
import org.processmining.plugins.petrinet.analysis.PNPluginSelectionPanel.AnalysisTypes;

/**
 * @author Arya Adriansyah, dfahland
 * @email a.adriansyah@tue.nl, d.fahland@tue.nl
 * @version Jun 4, 2010
 */
@Plugin(name = "Analyze Behavioral Property of Petri net", level = PluginLevel.PeerReviewed, categories = { PluginCategory.Analytics }, returnLabels = { "Analyzed net" }, returnTypes = { Petrinet.class }, parameterLabels = {
		"Net", "Marking", "Semantics", "Plugins to be executed" }, userAccessible = true, //
		mostSignificantResult = -1)
public class PNBehavioralAnalysis {

	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "A. Adriansyah, D. Fahland", email = "d.fahland@tue.nl", pack = "PNAnalysis")
	@PluginVariant(variantLabel = "Analyze Behavioral Properties of a Petri Net", requiredParameterLabels = { 0 })
	public Petrinet analyzePN(UIPluginContext context, Petrinet net) {
		// obtain its marking
		Marking m = null;
		try {
			InitialMarkingConnection con = context.getConnectionManager()
					.getFirstConnection(InitialMarkingConnection.class, context, net);
			m = (Marking) con.getObjectWithRole(InitialMarkingConnection.MARKING);
		} catch (ConnectionCannotBeObtained conProb) {
			JOptionPane.showMessageDialog(new JPanel(),
					"No behavioral analysis can be performed as the net does not have any corresponding initial marking");
			context.getFutureResult(0).cancel(true);
			return net;
		}
		analyzePNAssumeMarkingConnnection(context, net, m,
				PetrinetSemanticsFactory.regularPetrinetSemantics(Petrinet.class));
		return net;
	}

	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "A. Adriansyah, D. Fahland", email = "d.fahland@tue.nl", pack = "PNAnalysis")
	@PluginVariant(variantLabel = "Analyze Behavioral Properties of a Petri Net", requiredParameterLabels = { 0, 1 })
	public Petrinet analyzePN(UIPluginContext context, Petrinet net, Marking initialMarking) {
		// obtain its marking
		try {
			context.getConnectionManager().getFirstConnection(InitialMarkingConnection.class, context, net,
					initialMarking);
		} catch (ConnectionCannotBeObtained conProb) {
			JOptionPane.showMessageDialog(new JPanel(),
					"No behavioral analysis can be performed as the provided initial marking does not connected to the net");
			context.getFutureResult(0).cancel(true);
			return net;
		}
		analyzePNAssumeMarkingConnnection(context, net, initialMarking,
				PetrinetSemanticsFactory.regularPetrinetSemantics(Petrinet.class));
		return net;
	}

	@PluginVariant(variantLabel = "Analyze Behavioral Property of Petri Net", requiredParameterLabels = { 0, 1, 2 })
	public Petrinet analyzePNAssumeMarkingConnnection(UIPluginContext context, Petrinet net, Marking m,
			PetrinetSemantics sem) {
		// check if previous result exists
		Set<AnalysisTypes> available = new HashSet<AnalysisTypes>();

		for (AnalysisTypes type : PNPluginSelectionPanel.AnalysisTypes.values()) {
			try {
				if (type != AnalysisTypes.ALLBEHAVIORAL && type.isBehavioral()) {
					context.getConnectionManager().getFirstConnection(type.getConnectionType(), context, net, m, sem);
					available.add(type);
				}
			} catch (ConnectionCannotBeObtained e) {
				// Don't care
			}
		}

		Map<AnalysisTypes, Set<AnalysisTypes>> dependencies = DependencyFactory.getBehavioralPluginDependencies();

		PNPluginSelectionPanel panel = new PNPluginSelectionPanel(available, dependencies, "Select Analysis Plugins",
				"To select a plugin, click the name of the plugin on the list below. To deselect it, click again the name of the plugin. <br/>"
						+ "If a plug-in depends on other plugin(s), the plugins will automatically be selected.");

		InteractionResult result = context.showConfiguration("Behavioral Analysis Plug-in", panel);
		if (result != InteractionResult.CONTINUE) {
			context.getFutureResult(0).cancel(true);
			return net;
		}

		Set<AnalysisTypes> propertiesToBeCalculated = panel.getPropertiesToBeCalculated();

		if (propertiesToBeCalculated.size() > 0) {
			// update results
			// calculate result
			analyzeBehavioralProperty(context, net, m, sem, propertiesToBeCalculated.toArray(new AnalysisTypes[0]));

		} else {
			context.getFutureResult(0).cancel(true);
		}
		return net;
	}

	@PluginVariant(variantLabel = "Analyze Behavioral Property of Petri Net", requiredParameterLabels = { 0, 1, 2, 3 })
	public Petrinet analyzeBehavioralProperty(PluginContext context, Petrinet net, Marking m, PetrinetSemantics semantics,
			AnalysisTypes[] pluginToBeExecuted) {

		semantics.initialize(net.getTransitions(), new Marking(m));

		for (AnalysisTypes type : pluginToBeExecuted) {
			if (type != AnalysisTypes.ALLBEHAVIORAL) {
				try {
					context.tryToFindOrConstructFirstObject(type.getObjectType(), type.getConnectionType(),
							type.getRole(), net, m, semantics);
				} catch (ConnectionCannotBeObtained e) {
					context.log(e);
				}
			}
		}
		return net;
	}
}
