/***********************************************************
 * This software is part of the ProM package * http://www.processmining.org/ * *
 * Copyright (c) 2003-2008 TU/e Eindhoven * and is licensed under the * LGPL
 * License, Version 1.0 * by Eindhoven University of Technology * Department of
 * Information Systems * http://www.processmining.org * *
 ***********************************************************/

package org.processmining.plugins.petrinet.structuralanalysis;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import org.processmining.framework.connections.ConnectionCannotBeObtained;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.framework.util.collection.MultiSet;
import org.processmining.models.connections.petrinets.structural.AbstractInvariantMarkingConnection;
import org.processmining.models.connections.petrinets.structural.TComponentConnection;
import org.processmining.models.connections.petrinets.structural.TransitionInvariantConnection;
import org.processmining.models.graphbased.directed.petrinet.InhibitorNet;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.PetrinetGraph;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.analysis.TComponentSet;
import org.processmining.models.graphbased.directed.petrinet.analysis.TransitionInvariantSet;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.plugins.petrinet.structuralanalysis.invariants.TransitionInvariantCalculator;

import cern.colt.list.DoubleArrayList;
import cern.colt.list.IntArrayList;
import cern.colt.matrix.DoubleMatrix2D;

/**
 * Class to decompose petri net into its T-components (if there is any) this
 * implementation is similar to SComponentGenerator implementation
 * 
 * @author arya
 * @email arya.adriansyah@gmail.com
 * @version Nov 6, 2008
 */
@Plugin(name = "Identify T-Components", returnLabels = { "T-Components" }, returnTypes = { TComponentSet.class }, parameterLabels = {
		"Net", "Basis of transition Invariants" }, userAccessible = true)
public class TComponentGenerator {

	// variants with transition invariant as input
	@PluginVariant(variantLabel = "Identify T-Components", requiredParameterLabels = { 0, 1 })
	public TComponentSet calculateTComponentsPetriNet(PluginContext context, Petrinet net,
			TransitionInvariantSet invMarking) throws Exception {
		return calculateTComponentsInternal(context, net, invMarking);
	}

	@PluginVariant(variantLabel = "Identify T-Components", requiredParameterLabels = { 0, 1 })
	public TComponentSet calculateSComponentsInhibitorNet(PluginContext context, InhibitorNet net,
			TransitionInvariantSet invMarking) throws Exception {
		return calculateTComponentsInternal(context, net, invMarking);
	}

	// variants with no transition invariant as input
	@PluginVariant(variantLabel = "Identify T-Components", requiredParameterLabels = { 0 })
	public TComponentSet calculateSComponentsPetriNet(PluginContext context, Petrinet net) throws Exception {
		return calculateTComponentsInternalWithInvariantAddition(context, net);
	}

	@PluginVariant(variantLabel = "Identify T-Components", requiredParameterLabels = { 0 })
	public TComponentSet calculateSComponentsInhibitorNet(PluginContext context, InhibitorNet net) throws Exception {
		return calculateTComponentsInternalWithInvariantAddition(context, net);
	}

	/**
	 * This method generate invariants or taking an existing invariants from
	 * context in order to continue calculation of T Component
	 * 
	 * @param context
	 *            context of the net
	 * @param net
	 *            net to be calculated
	 * @return TComponentSet set of T-Component in the net
	 * @throws Exception
	 */
	private TComponentSet calculateTComponentsInternalWithInvariantAddition(PluginContext context, PetrinetGraph net)
			throws ConnectionCannotBeObtained {
		TransitionInvariantSet invMarking = null;

		invMarking = context.tryToFindOrConstructFirstObject(TransitionInvariantSet.class,
				TransitionInvariantConnection.class, AbstractInvariantMarkingConnection.INVARIANTMARKING, net);

		return calculateTComponentsInternal(context, net, invMarking);
	}

	/**
	 * Main method to calculate T-Components
	 * 
	 * @param context
	 *            context of the net
	 * @param net
	 *            net to be calculated
	 * @param invMarking
	 *            Transition invariant of the corresponding net
	 * @return TComponentSet set of T-Component in the net
	 * @throws Exception
	 */
	private TComponentSet calculateTComponentsInternal(PluginContext context, PetrinetGraph net,
			TransitionInvariantSet invMarking) throws ConnectionCannotBeObtained {
		// check connectivity between invariant and net
		context.getConnectionManager()
				.getFirstConnection(TransitionInvariantConnection.class, context, net, invMarking);

		// calculate T-Component
		TComponentSet nodeMarking = calculateTComponents(context, net, invMarking);

		// add connection for t-component
		context.addConnection(new TComponentConnection(net, nodeMarking));
		context.getFutureResult(0).setLabel("T-Components of " + net.getLabel());

		return nodeMarking;
	}

	/**
	 * Main method to calculate T-Component (with context)
	 * 
	 * @param net
	 *            PetrinetGraph with its T-components to be calculate
	 * @param invMarking
	 *            Invariant to be used in calculation.
	 * @return TComponentSet set of T-Component in the net
	 */
	private TComponentSet calculateTComponents(PluginContext context, PetrinetGraph net,
			TransitionInvariantSet invMarking) {
		// initialization
		TComponentSet nodeMarking = new TComponentSet(); // to store final result

		if (invMarking == null) {
			// calculate place invariants
			TransitionInvariantCalculator calculator = new TransitionInvariantCalculator();
			invMarking = calculator.calculate(context, net);
		}

		// references 
		DoubleMatrix2D incidenceMatrix = IncidenceMatrixFactory.getIncidenceMatrix(context, net);
		if (context != null && context.getProgress().isCancelled()) {
			return null;
		}

		ArrayList<Place> placeList = new ArrayList<Place>(net.getPlaces());
		ArrayList<Transition> transitionList = new ArrayList<Transition>(net.getTransitions());

		// for each transition invariants filter only the one with 1 or 0 as its member
		invariantLoop: for (MultiSet<Transition> set : invMarking) {
			
			if (context != null && context.getProgress().isCancelled()) {
				return null;
			}

			// for one set of invariant, check each element
			Set<Integer> transitionIndex = new HashSet<Integer>();
			for (PetrinetNode node : set) {
				if (set.occurrences(node) == 1) {
					transitionIndex.add(transitionList.indexOf(node));
				} else {
					continue invariantLoop;
				}
			}

			// until here, we have an invariant which only consists of 0 and 1

			// iterate the transitions, generate array of integer to dice incidenceMatrix
			int[] transitions = new int[transitionIndex.size()];
			Iterator<Integer> it = transitionIndex.iterator();
			int counter = 0;
			while (it.hasNext()) {
				transitions[counter] = it.next();
				counter++;
			}
			// iterate all of places, generate array of integer to dice incidenceMatrix
			int[] places = new int[placeList.size()];
			for (int i = 0; i < placeList.size(); i++) {
				places[i] = i;
			}

			// dice incidence matrix so that it only include necessary place and transitions
			DoubleMatrix2D tempIncidenceMatrix = incidenceMatrix.viewSelection(places, transitions);

			Set<Integer> placeIndex = new HashSet<Integer>(); // to store result for transition

			// for each rows on the diced incidence matrix, there can only be 2 nonzero value with the sum of 0
			IntArrayList tempPlace = new IntArrayList();
			DoubleArrayList tempPlaceValue = new DoubleArrayList();
			for (int i = 0; i < tempIncidenceMatrix.rows(); i++) {
				tempIncidenceMatrix.viewRow(i).getNonZeros(tempPlace, tempPlaceValue);

				// as only 1 ingoing and 1 outgoing arc is permitted, number of element should be 2
				// as there can only be 1 ingoing and 1 outgoing, addition should be 0
				if ((tempPlace.size() == 2)
						&& (Double.compare(0.0, tempPlaceValue.get(0) + tempPlaceValue.get(1)) == 0)) {
					// add to transitionIndex
					placeIndex.add(i);
				} else if (tempPlace.size() != 0) {
					// this invariant has invalid transition inclusion. Continue to other invariants
					continue invariantLoop;
				}
			}

			// until here, we have our T-components, add all transition and places as a T-component
			SortedSet<PetrinetNode> result = new TreeSet<PetrinetNode>();
			// add all places
			for (int tempIndex : transitionIndex) {
				result.add(transitionList.get(tempIndex));
			}
			// add transition corresponds to each place
			for (int tempIndex : placeIndex) {
				result.add(placeList.get(tempIndex));
			}
			// add to final set
			nodeMarking.add(result);
		}

		return nodeMarking;
	}

}
