/**
 * 
 */
package org.processmining.plugins.petrinet.structuralanalysis.invariants;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Set;
import java.util.SortedSet;

import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.framework.plugin.events.Logger.MessageLevel;
import org.processmining.framework.util.collection.SortedMultiSet;
import org.processmining.framework.util.collection.TreeMultiSet;
import org.processmining.models.connections.petrinets.structural.PlaceInvariantConnection;
import org.processmining.models.graphbased.directed.petrinet.InhibitorNet;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetGraph;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.analysis.PlaceInvariantSet;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.plugins.petrinet.structuralanalysis.IncidenceMatrixFactory;

import cern.colt.matrix.DoubleMatrix2D;

/**
 * This class represents a plugin to calculate basis of semi-positive place
 * invariants of petri net Modified from PlaceInvariantCalculator class
 * implemented in ProM 5
 * 
 * @author arya
 * @email arya.adriansyah@gmail.com
 * @version Oct 24, 2008
 */
@Plugin(name = "Construct Basis of Semi-Positive Place Invariants", returnLabels = { "Basis of Place Invariants" }, returnTypes = { PlaceInvariantSet.class }, parameterLabels = { "Net" }, userAccessible = true)
public class PlaceInvariantCalculator {

	@PluginVariant(variantLabel = "Construct Basis of Semi-Positive Place Invariants", requiredParameterLabels = { 0 })
	public PlaceInvariantSet calculatePlaceInvariant(PluginContext context, Petrinet net) {
		return calculatePlaceInvariantInternal(context, net);
	}

	@PluginVariant(variantLabel = "Construct Basis of Semi-Positive Place Invariants", requiredParameterLabels = { 0 })
	public PlaceInvariantSet calculatePlaceInvariant(PluginContext context, InhibitorNet net) {
		return calculatePlaceInvariantInternal(context, net);
	}

	/**
	 * Main method to calculate place invariant
	 * 
	 * @param context
	 *            context of the net
	 * @param net
	 *            net to be calculated
	 * @return PlaceInvariantSet set of place invariants
	 */
	private PlaceInvariantSet calculatePlaceInvariantInternal(PluginContext context, PetrinetGraph net) {
		context.getFutureResult(0).setLabel("Basis of semi-positive place invariants of " + net.getLabel());

		// get incidence matrix for the net
		DoubleMatrix2D incidenceMatrix = IncidenceMatrixFactory.getTransposedIncidenceMatrix(context, net);

		context.log("INCIDENCE MATRIX", MessageLevel.DEBUG);
		context.log("number of rows : " + incidenceMatrix.rows(), MessageLevel.DEBUG);
		context.log("number of cols : " + incidenceMatrix.columns(), MessageLevel.DEBUG);

		StringBuffer tempOut = new StringBuffer();
		for (int i = 0; i < incidenceMatrix.rows(); i++) {
			tempOut.delete(0, tempOut.length());
			for (int j = 0; j < incidenceMatrix.columns(); j++) {
				tempOut.append(" ");
				tempOut.append(incidenceMatrix.get(i, j));
			}
			context.log(tempOut.toString(), MessageLevel.DEBUG);
		}

		// calculate invariants
		PlaceInvariantSet pinv = calculate(context, net); // petrinet is always
		// short-circuited in ProM 5

		// add connection for invariant marking
		context.addConnection(new PlaceInvariantConnection(net, pinv));
		return pinv;
	}

	/**
	 * Calculation of place invariant marking This method is public in order to
	 * enable invariant calculation as intermediate step
	 * 
	 * @param net
	 *            net to be calculated
	 * @return
	 */
	public PlaceInvariantSet calculate(PluginContext context, PetrinetGraph net) {
		// calculate invariant, store it in inv
		PlaceInvariantSet inv = new PlaceInvariantSet("Basis of semi-positive place invariants of " + net.getLabel());
		DoubleMatrix2D m = InvariantCalculator.calculate(context,
				IncidenceMatrixFactory.getTransposedIncidenceMatrix(context, net).viewDice());

		if (context != null && context.getProgress().isCancelled()) {
			return null;
		}

		// create a reference table to check edges
		Set<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> edges = net.getEdges();
		Iterator<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> iterator = edges.iterator();

		HashMap<PetrinetNode, PetrinetNode> reference = new HashMap<PetrinetNode, PetrinetNode>();
		while (iterator.hasNext()) {
			PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> edge = iterator.next();
			reference.put(edge.getSource(), edge.getTarget());
		}

		// calculate
		for (int i = 0; i < m.rows(); i++) {

			if (context != null && context.getProgress().isCancelled()) {
				return null;
			}

			// For each row, make a set over the places
			SortedMultiSet<Place> b = new TreeMultiSet<Place>();

			int j = 0;
			for (Place p : net.getPlaces()) {
				for (int k = 0; k < m.get(i, j); k++) {
					b.add(p);
				}
				j++;
			}

			// Arya: apparently, the result may not be the basis yet. Need to
			// divide the number of elements with their greatest common divisor
			// (GCD)
			SortedSet<Place> base = b.baseSet();
			int[] occurrences = new int[base.size()];
			int counter = 0;
			for (Place p : base) {
				occurrences[counter] = b.occurrences(p);
				counter++;
			}
			int gcd = InvariantCalculator.getGCDs(occurrences);
			if (gcd > 1) {
				// divide all occurrences
				SortedMultiSet<Place> bprime = new TreeMultiSet<Place>();
				for (Place p : base) {
					bprime.add(p, b.occurrences(p) / gcd);
				}
				inv.add(bprime);
			} else {
				inv.add(b);
			}
			// end of addition
		}

		return inv;
	}

}
